/*
 * factory.h
 *
 *  Created on: Dec 17, 2010
 *      Author: hans
 */

#ifndef FACTORY_H_
#define FACTORY_H_

#include <string>
#include <exception>
#include "binary_matrix.h"
#include "large_sliced_coo.h"
#include "sliced_coo.h"
#include "small_sliced_coo.h"
#include "xl_sliced_coo.h"
#include "sliced_ellpack.h"
#include "ad_sliced_coo.h"

namespace cuda_bwc{

namespace matrix_factory {

class UnknownMatrixFormatException: public std::exception
{
public:
	UnknownMatrixFormatException(const std::string &msg) : std::exception(), msg(msg){};
	~UnknownMatrixFormatException() throw(){};
	virtual const char* what() const throw(){
		return msg.c_str();
	}
private:
	std::string msg;
};


const std::string BIT_BLOCK("bitblock");
const std::string LARGE_SLICED_COO("lscoo");
const std::string SMALL_SLICED_COO("sscoo");
const std::string XL_SLICED_COO("xlscoo");
const std::string SLICED_ELLPACK("sle");

const std::string SLICED_COO_256("256scoo");
const std::string SLICED_COO_128("128scoo");
const std::string SLICED_COO_64("64scoo");
const std::string SLICED_COO_32("32scoo");
const std::string SLICED_COO_16("16scoo");
const std::string SLICED_COO_8("8scoo");
const std::string SLICED_COO_4("4scoo");
const std::string SLICED_COO_2("2scoo");
const std::string SLICED_COO_1("1scoo");

const std::string AD_SLICED_COO_2("2ascoo");
const std::string AD_SLICED_COO_4("4ascoo");
const std::string AD_SLICED_COO_8("8ascoo");

const std::string SMALL_SLICED_COO_64("64sscoo");
const std::string SMALL_SLICED_COO_128("128sscoo");
const std::string SMALL_SLICED_COO_256("256sscoo");


template <typename T>
BinaryMatrix<T>* getMatrixObject(const std::string& format) {
//	if(format == BIT_BLOCK){
//		return new BitBlock<T>();
//	} else if(format == LARGE_SLICED_COO){
//		return new LargeSlicedCoo<T>();
	/*} else */if(format == SMALL_SLICED_COO){
		return new SmallSlicedCoo<T,512>();
    } else if(format == SMALL_SLICED_COO_256){
        return new SmallSlicedCoo<T,256>();
    } else if (format == SMALL_SLICED_COO_128) {
        return new SmallSlicedCoo<T,128>();
    } else if (format == SMALL_SLICED_COO_64) {
        return new SmallSlicedCoo<T,64>();
	} else if(format == XL_SLICED_COO){
		return new XLSlicedCoo<T>();
    } else if (format ==SLICED_COO_64) {
        return new SlicedCoo<T, 64>();
    } else if (format == SLICED_COO_128) {
        return new SlicedCoo<T, 128>();
    } else if (format == SLICED_COO_256) {
        return new SlicedCoo<T, 256>();
    } else if (format == SLICED_COO_32) {
        return new SlicedCoo<T,32>();
    } else if(format == SLICED_COO_16) {
        return new SlicedCoo<T, 16>();
    } else if(format == SLICED_COO_8) {
        return new SlicedCoo<T,8>();
    } else if (format == SLICED_COO_4) {
        return new SlicedCoo<T,4>();
    } else if (format == SLICED_COO_2) {
        return new SlicedCoo<T,2>();
    } else if (format == SLICED_COO_1) {
        return new SlicedCoo<T,1>();
//    } else if (format == AD_SLICED_COO_2) {
//        return new AdSlicedCoo<T,2>();
//     } else if (format == AD_SLICED_COO_4) {
//        return new AdSlicedCoo<T,4>();
//     } else if (format == AD_SLICED_COO_8) {
//        return new AdSlicedCoo<T,8>();
	} else if(format == SLICED_ELLPACK){
        return new SlicedEllpack<T>();
	} else{
		std::stringstream ss;
		ss << "Format " << format << " is not recognized.";
		throw UnknownMatrixFormatException(ss.str());
	}
    //return NULL;
}

}

template class BinaryMatrix<float>;
template class BinaryMatrix<double>;

}


#endif /* FACTORY_H_ */
